

const Cc = Components.classes;
const Ci = Components.interfaces;
const Cu = Components.utils;

Cu.import("resource://gre/modules/Services.jsm");
Cu.import("resource://gre/modules/NetUtil.jsm");

function isNativeUI() {
    let appInfo = Cc["@mozilla.org/xre/app-info;1"].getService(Ci.nsIXULAppInfo);
    return (appInfo.ID == "{aa3c5121-dab2-40e2-81ca-7ea25febc110}");
}

var menuId;


function log(msg) {
    Services.console.logStringMessage(msg);
}
function dump(msg) {
    Services.console.logStringMessage(msg);
}

function writeString(stream, string) {
    stream.write(string, string.length);
}



/* MirrorServer implements a very basic HTTP server that only serves a single
 * screenshot of the active tab content on a non-standard port. Google Cast
 * supports loading content from a URL, but a screenshot data: URI is too large
 * to send. Instead, we send a unique URL which is served by the MirrorServer.
 */
function MirrorServer(aContent) {
    this.server = Cc["@mozilla.org/network/server-socket;1"].createInstance(Ci.nsIServerSocket);
    this.server.init(9191, false, 5);
    this.server.asyncListen(this);
    log("Opened socket on " + this.server.port);

    this.input = null;
    this.output = null;
    this.content = aContent;
}

MirrorServer.prototype = {
    get port() {
        return this.server.port;
    },

    stop: function() {
        if (this.input)
            this.input.close();
        if (this.output)
            this.output.close();
        this.server.close();
    },

    setContent: function(aContent) {
        this.content = aContent;
    },

    onSocketAccepted: function(aServerSocket, aClientSocket) {
        log("Accepted connection on " + aClientSocket.host + ":" + aClientSocket.port);

        this.input = aClientSocket.openInputStream(Ci.nsITransport.OPEN_BLOCKING, 100, 0).QueryInterface(Ci.nsIAsyncInputStream);
        this.output = aClientSocket.openOutputStream(Ci.nsITransport.OPEN_BLOCKING, 0, 0);
        this.input.asyncWait(this, 0, 0, Services.tm.mainThread);
    },

    sendResponse: function() {
        let binaryOutput = Cc["@mozilla.org/binaryoutputstream;1"].createInstance(Ci.nsIObjectOutputStream);
        binaryOutput.setOutputStream(this.output);

        let binaryStream = Cc["@mozilla.org/binaryinputstream;1"].createInstance(Ci.nsIBinaryInputStream);

        binaryStream.setInputStream(this.content.stream);
        let data = this._makeHeader(0, this.content.type);
        writeString(binaryOutput, data);
        try {
            let avail = binaryStream.available();
            do {
                writeString(binaryOutput, avail + "\r\n");
                dump("writing  " + avail + " bytes");
                //binaryOutput.writeFrom(binaryStream, avail);
                binaryOutput.writeByteArray(binaryStream.readByteArray(avail), avail);
                writeString(binaryOutput, "\r\n");
                avail = binaryStream.available();
            } while(avail > 0)
        } catch(ex) { dump("exception reading stream" + ex); } // stream will throw when it closes
        dump("reached the end of the stream, closing");
        writeString(binaryOutput, "0\r\n\r\n");
        binaryOutput.close();

    },

    onStopListening : function(aSocket, aStatus) {
    },

    onInputStreamReady : function(aInput) {
        let stream = Cc["@mozilla.org/scriptableinputstream;1"].createInstance(Ci.nsIScriptableInputStream);
        stream.init(this.input);
        let request = "";
        try {
            while (stream.available())
            request = request + stream.read(512);
        } catch (e) {
            log("Received (end): " + request);
        }
        log("Received: " + request);
        log(request.substring(4, 10));
        let window = this.content.window;
        if (request.substring(4, 10) == "/image") {
            let type = "image/png";
            let extra = getContentAsImage(window, window.BrowserApp.selectedTab, function(aStream) {
                gMirror.setContent({ stream: aStream, type: type, window: window});
                gMirror.sendResponse()
            }.bind(this), type);
        } else {
            let type = "video/webm";
            let obj = {};
            let extra = getContentAsVideo(window, window.BrowserApp.selectedTab, function(aStream) {
                gMirror.setContent({ stream: aStream, type: type, window: window});
                gMirror.sendResponse()
            }.bind(obj), type);
        }
        this.input.close();
    },

    _makeHeader: function(aLength, aType) {
        let headers = [];
        headers.push("HTTP/1.1 200 OK");
        headers.push("Server: FirefoxCast");
        headers.push("Transfer-Encoding: chunked");
        headers.push("Content-Type: " + aType);
        headers.push("\n");
        return headers.join("\r\n");
    }
}



function getContentAsVideo(window, tab, aCallback, aType) {
   window.navigator.mozGetTabStream(0, 100, 400, 300, function(stream) {
        for (let i in stream) {
            dump(i +": " + stream[i]);
        }
        let tracks = stream.getVideoTracks();
        for (let i in tracks) {
            dump(i +": " + tracks[i]);
        }
        let track = tracks[0];
        for (let i in track) {
            dump(i +": " + track[i]);
        }

       dump(stream);

       //let localStream = stream.QueryInterface(Ci.nsIInterfaceRequestor).getInterface(Ci.nsIDOMLocalMediaStream);
       dump("got a stream: " + stream);
       let ios = stream.getEncodedStream(aType);
       dump("got an input stream: " + ios);
        //var url = window.URL.createObjectURL(stream);
        aCallback(ios);
    }, function (error) { dump(error);});
}

function getContentAsImage(window, tab, aCallback, aType) {
    let viewport = tab.getViewport();
    let left = viewport.cssX;
    let top = viewport.cssY;
    let width = Math.round(viewport.cssWidth);
    let height = Math.round(viewport.cssHeight);

    let canvas = tab.window.document.createElementNS("http://www.w3.org/1999/xhtml", "canvas");
    canvas.width = width;
    canvas.height = height;

    let context = canvas.getContext("2d");
    let flags = context.DRAWWINDOW_DRAW_CARET | context.DRAWWINDOW_DO_NOT_FLUSH | context.DRAWWINDOW_USE_WIDGET_LAYERS;
    context.drawWindow(tab.window, left, top, width, height, "#fff", flags);
    canvas.mozFetchAsStream(aCallback, aType);

    // Free the memory
    canvas.width = 0;
    canvas.height = 0;
    canvas = null;

    return { width: width, height: height };
}

let gMirror;

function streamTab(window) {
    gMirror = new MirrorServer({ stream: null, type: null, window: window});
}

function loadIntoWindow(window) {
    if (!window)
        return;
    menuId = window.NativeWindow.menu.add("Stream Tab", null, function() { streamTab(window); });
}

function unloadFromWindow(window) {
    if (!window)
        return;
    window.NativeWindow.menu.remove(menuId);
}

var windowListener = {
    onOpenWindow: function(aWindow) {
        // Wait for the window to finish loading
        let domWindow = aWindow.QueryInterface(Ci.nsIInterfaceRequestor).getInterface(Ci.nsIDOMWindowInternal || Ci.nsIDOMWindow);
        domWindow.addEventListener("load", function onLoad() {
            domWindow.removeEventListener("load", onLoad, false);
            loadIntoWindow(domWindow);
        }, false);
    },

    onCloseWindow: function(aWindow) {},
    onWindowTitleChange: function(aWindow, aTitle) {}
};

function startup(aData, aReason) {
    let wm = Cc["@mozilla.org/appshell/window-mediator;1"].getService(Ci.nsIWindowMediator);

    // Load into any existing windows
    let windows = wm.getEnumerator("navigator:browser");
    while (windows.hasMoreElements()) {
        let domWindow = windows.getNext().QueryInterface(Ci.nsIDOMWindow);
        loadIntoWindow(domWindow);
    }

    // Load into any new windows
    wm.addListener(windowListener);
}

function shutdown(aData, aReason) {
    if (gMirror != null) {
        gMirror.stop();
        gMirror = null;
    }

    // When the application is shutting down we normally don't have to clean
    // up any UI changes made
    if (aReason == APP_SHUTDOWN)
        return;

    let wm = Cc["@mozilla.org/appshell/window-mediator;1"].getService(Ci.nsIWindowMediator);

    // Stop listening for new windows
    wm.removeListener(windowListener);

    // Unload from any existing windows
    let windows = wm.getEnumerator("navigator:browser");
    while (windows.hasMoreElements()) {
        let domWindow = windows.getNext().QueryInterface(Ci.nsIDOMWindow);
        unloadFromWindow(domWindow);
    }
}

function install(aData, aReason) {}
function uninstall(aData, aReason) {}
